/*
 * qemu_hugepages.c: QEMU memory control
 *
 * Copyright (c) 2020 Oracle and/or its affiliates
 *
 * Autonomous memoryBacking nr_hugepage reservations under control of libvirt
 */

#include <config.h>
#include <sys/mman.h>

#include "qemu_hugepages.h"

#include "viralloc.h"
#include "virlog.h"
#include "virnuma.h"
#include "virhostmem.h"

#define VIR_FROM_THIS VIR_FROM_QEMU

VIR_LOG_INIT("qemu.qemu_memory");

#if defined(ENABLE_EXADATA)
# define HP_TARGET (GuestIsNUMA ? cell : -1)


/* _qemuRetrySetPagePoolSize
 *
 * Controlling the Linux Page Pool Size to target 'host' arranged 2MiB or 1GiB
 * acquisition by virNumaSetPagePoolSize is surprising erratic when operated per SysFS
 * NUMA node-specific entries.  Requests are interrupted midstream and in need of retry
 * till the operation finishes.   Returning an error to its callee results in a fatality to
 * the arranged 'guest' domain toolstack, so this routine provides a mechanism to retry by
 * timeout, which propagates to the best operational demand.  Low demand steers towards
 * the minimum timeout, which high demand directs to a long timeout.  These timeouts,
 * if in effect, add at max a couple of seconds to a NUMA node memory acquisition
 * or release or for interactive individual DIMM HOT-PLUG and HOT-UNPLUG operations.
 * If an unusual event happens, we report the system error but never inform the callee
 * about the fatality. Hence there is _NO_ Errno anyways, as such would result in a full
 * libvirt stack caveat.
 *
 * WTH_RESOVE: This needs future attention; errors need to be dealt with.
 */
static void
_qemuRetrySetPagePoolSize(virQEMUDriverConfigPtr cfg,
                          virDomainDefPtr def,
                          size_t cell,
                          int target,
                          size_t Hugepagesize,
                          unsigned long long whence,
                          bool append)
{
    int err = -1;
    size_t retry = cfg->exadataConfig->retryHugePages;
    static size_t timeout;

    if (!timeout)
        timeout = cfg->exadataConfig->timeoutHugePages;
    else if (timeout > cfg->exadataConfig->timeoutHugePages)
        timeout -= cfg->exadataConfig->timeoutHugePages;

    do {
        if (retry < cfg->exadataConfig->retryHugePages)
            timeout += cfg->exadataConfig->timeoutHugePages;
        usleep(timeout);
    } while (retry-- && (err = virNumaSetPagePoolSize(target, Hugepagesize, whence, append)) < 0);

    /* There's a bad condition.  Make a libvirt report but keep it silent. */
    if (err)
        virReportError(VIR_ERR_SYSTEM_ERROR,
                _("Failed to reset the reservation "
                  "of %lld pages under node %zu for domain '%s'"),
                whence, cell, def->name);

    return;
}


static int
_qemuGetNUMAPagesInfo(int node,
                      qemuHostNUMACellPageInfoPtr *pageinfo,
                      size_t *npageinfo)
{
    int ret = -1;
    unsigned int *pages_size = NULL;
    unsigned long long *pages_avail = NULL;
    unsigned long long *pages_free = NULL;
    size_t npages, i;

    if (virNumaGetPages(node, &pages_size,
                &pages_avail, &pages_free, &npages) < 0)
        goto cleanup;

    if (VIR_ALLOC_N(*pageinfo, npages) < 0)
        goto cleanup;

    *npageinfo = npages;

    for (i = 0; i < npages; i++) {
        (*pageinfo)[i].size = pages_size[i];
        (*pageinfo)[i].avail = pages_avail[i];
        (*pageinfo)[i].free  = pages_free[i];
    }

    ret = 0;

 cleanup:
    VIR_FREE(pages_free);
    VIR_FREE(pages_avail);
    VIR_FREE(pages_size);
    return ret;
}


int
qemuMemAcquireHostHugepages(virQEMUDriverPtr driver,
                            virDomainDefPtr def,
                            virQEMUDriverConfigPtr cfg,
                            unsigned long long size,
                            int targetNode)
{
    if (geteuid() == 0 &&
        cfg->exadataConfig &&
        cfg->exadataConfig->enaAutonomousHugepages &&
        cfg->exadataConfig->pctHugePages &&
        def->mem.nhugepages &&
        def->mem.hugepages && !def->mem.hugepages[0].nodemask) {

        qemuHostNUMACellPageInfoPtr pageinfo = NULL;
        qemuGuestNUMACellReqPtr nodepages = NULL;
        int err = -1;
        size_t startcell = 0, endcell;
        size_t cell, ncells;
        size_t npageinfo;
        size_t Hugepagesize = 2048L;
        bool GuestIsNUMA = true;
        unsigned long long GuestNodeSize = 0, HostNodeFree = 0;
        unsigned long long memorybytes;

        /* Calculate the maximum percentage setting for a guest's
         * memory relative to the host's total available memory. */
        if (virHostMemGetInfo(&memorybytes, NULL) < 0)
            goto cleanup;

        if (virDomainDefGetMemoryTotal(def) > \
            ((cfg->exadataConfig->pctHugePages * (memorybytes / 1024L)) / 100L)) {
            virReportError(VIR_ERR_OPERATION_INVALID,
                           _("Domain '%s' configured memoryBacking of %lld KiB, exceeds the host "
                             "%d%% restricted hugepage acquisition. Max allowed is %lld KiB."),
                           def->name, virDomainDefGetMemoryTotal(def),
                           cfg->exadataConfig->pctHugePages,
                           ((cfg->exadataConfig->pctHugePages * (memorybytes / 1024L)) / 100L));
            goto cleanup;
        }

        if (cfg->exadataConfig->spareHugePages < MIN_HUGEPAGE_SPAREPAGES) {
            cfg->exadataConfig->spareHugePages = MIN_HUGEPAGE_SPAREPAGES;

            VIR_WARN("Adjusting the number of hugepages to keep as spare to "
                     "%d 2MiB pages per NUMA node", MIN_HUGEPAGE_SPAREPAGES);
        }

        /* Is this a HOT-PLUG? */
        if (targetNode >= 0) {

            /* Will it fit the available slots? If so, then don't cause
             * an exception.  The libvirt domain code takes a reversed
             * approach towards HOT-PLUG memory, checking its available
             * slots postmortem. */
            if (def->nmems >= def->mem.memory_slots) {
                err = 0;
                goto cleanup;
            }

            /* HOT-PLUG goes by single NUMA nodes. */
            ncells = 1;
            startcell = targetNode;
            endcell = targetNode + ncells;
        } else {

            /* Get the vNUMA node count. */
            ncells = virDomainNumaGetNodeCount(def->numa);

            /* If none, there's always a NUMA node0! */
            if (!ncells) {
                ncells = 1;
                GuestIsNUMA = false;
            }
            endcell = ncells;
        }

        if (VIR_ALLOC_N(nodepages, endcell) < 0)
            goto cleanup;

        /* If the guest configured a specific hugepage setting,
         * overrule the default 2MiB setting. */
        if (def->mem.hugepages[0].size)
            Hugepagesize = def->mem.hugepages[0].size;

        /* Now lock all current and future pages,
         * preventing them from being paged */
        if ((driver->libvirtd_mlockall == false) &&
            cfg->exadataConfig->lckLibvirtdMlockall) {
                if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
                    virReportSystemError(errno, "%s",
                                         _("Failed to lock libvirtd memory"));
                    goto cleanup;
                }
                driver->libvirtd_mlockall = true;
        }

        /* LOCK domain hugepage acquisition. */
        virMutexLock(&driver->hugepage_lock);

        for (cell = startcell; cell < endcell; cell++) {

            if (_qemuGetNUMAPagesInfo(HP_TARGET, &pageinfo, &npageinfo) < 0)
                goto error;

            if (GuestIsNUMA) {
                if (size)
                    GuestNodeSize = size / ncells;
                else
                    GuestNodeSize = virDomainNumaGetNodeMemorySize(def->numa, cell);
            } else {
                GuestNodeSize = virDomainDefGetMemoryTotal(def);
            }

            HostNodeFree = (4L * (pageinfo[PAGE_4K].free));
            if (HostNodeFree < (Hugepagesize * cfg->exadataConfig->spareHugePages)) {
                virReportError(VIR_ERR_OPERATION_INVALID,
                               _("Insufficient memory to preserve host NUMA spare memory: "
                                 "node %zu spare pages, spare %lld, reserved %ld"),
                               cell, HostNodeFree, (Hugepagesize * cfg->exadataConfig->spareHugePages));
                goto error;
            }
            HostNodeFree -= (Hugepagesize * cfg->exadataConfig->spareHugePages);

            if (GuestNodeSize >= HostNodeFree) {
                virReportError(VIR_ERR_OPERATION_INVALID,
                               _("Insufficient memory available under NUMA "
                                 "node %zu, available %lld KiB, needed %lld KiB"),
                               cell, HostNodeFree, GuestNodeSize);
                goto error;
            }

            nodepages[cell].nr_pages = GuestNodeSize;
            nodepages[cell].nr_free  = HostNodeFree;
        }

        /* We still exist, so there's a fit! */
        for (cell = startcell; cell < endcell; cell++) {

            GuestNodeSize = nodepages[cell].nr_pages;
            HostNodeFree = nodepages[cell].nr_free;

            if (virNumaSetPagePoolSize(HP_TARGET, Hugepagesize,
                        GuestNodeSize / Hugepagesize, true) < 0) {

                virReportError(VIR_ERR_OPERATION_INVALID,
                               _("Failed to acquire %lld required %zu-KiB "
                                 "pages under NUMA node %zu"),
                              GuestNodeSize / Hugepagesize, Hugepagesize, cell);

                /* Restore (reset) the last functional page sets. */
                do {
                    _qemuRetrySetPagePoolSize(cfg, def, cell,
                            HP_TARGET, Hugepagesize, pageinfo[PAGE_2M].avail, false);

                } while (targetNode < 0 && cell--);

                goto error;
            }
        }

        /* Mark the hugepage acquisition perimeter. */
        def->acquiredHugepages = true;
        err = 0;

 error:
        /* UNLOCK domain hugepage acquisition. */
        virMutexUnlock(&driver->hugepage_lock);

 cleanup:
        VIR_FREE(pageinfo);
        VIR_FREE(nodepages);

        return err;
    }

    return 0;
}

int
qemuMemReleaseHostHugepages(virQEMUDriverPtr driver,
                            virDomainDefPtr def,
                            virDomainMemoryDefPtr mem)
{
    VIR_AUTOUNREF(virQEMUDriverConfigPtr) cfg = virQEMUDriverGetConfig(driver);

    if (geteuid() == 0 &&
        cfg->exadataConfig &&
        cfg->exadataConfig->enaAutonomousHugepages &&
        def->mem.nhugepages &&
        def->mem.hugepages && !def->mem.hugepages[0].nodemask) {

        int err = -1;
        bool GuestIsNUMA = true;
        size_t Hugepagesize = 2048L;
        size_t npageinfo;
        size_t i;
        size_t startcell = 0, endcell;
        size_t cell, ncells;
        unsigned long long  GuestNodeSize = 0;
        qemuHostNUMACellPageInfoPtr pageinfo = NULL;

        /* If this is a HOT-UNPLUG assure it is within a
         * valid slot range. */
        if (mem && (def->nmems >= def->mem.memory_slots))
            return 0;

        ncells = virDomainNumaGetNodeCount(def->numa);
        if (!ncells) {
            /* There's always a NUMA node0. */
            ncells = 1;
            GuestIsNUMA = false;
            GuestNodeSize = virDomainDefGetMemoryTotal(def) / Hugepagesize;
        }
        endcell = ncells;

        /* If the guest configured a specific hugepage setting,
         * overrule the default 2MiB setting. */
        if (def->mem.hugepages[0].size)
            Hugepagesize = def->mem.hugepages[0].size;

        /* LOCK domain hugepage acquisition. */
        virMutexLock(&driver->hugepage_lock);

        if (def->acquiredHugepages) {

            /* Is this an individual memory hot-unplug request? */
            if (mem) {
                cell = mem->targetNode;
                if (_qemuGetNUMAPagesInfo(cell, &pageinfo, &npageinfo) < 0)
                    goto error;

                GuestNodeSize = mem->size / Hugepagesize;

                _qemuRetrySetPagePoolSize(cfg, def, cell,
                        HP_TARGET, Hugepagesize, pageinfo[PAGE_2M].avail - GuestNodeSize, false);
            } else {

                /* Guest domain cleanup, tear down all vNUMA node acquired memory. */
                for (cell = startcell; cell < endcell; cell++) {

                    if (_qemuGetNUMAPagesInfo(HP_TARGET, &pageinfo, &npageinfo) < 0)
                        goto error;

                    if (!GuestNodeSize)
                        GuestNodeSize = virDomainNumaGetNodeMemorySize(def->numa, cell) / Hugepagesize;

                    _qemuRetrySetPagePoolSize(cfg, def, cell,
                            HP_TARGET, Hugepagesize, pageinfo[PAGE_2M].avail - GuestNodeSize, false);
                }

                /* Guest domain cleanup, tear down all guest domain hot-plugged
                 * acquired memory. */
                for (i = 0; i < def->nmems; i++) {
                    virDomainMemoryDefPtr mdef = def->mems[i];

                    if (mdef) {
                        cell = mdef->targetNode;

                        if (_qemuGetNUMAPagesInfo(cell, &pageinfo, &npageinfo) < 0)
                            goto error;

                        GuestNodeSize = mdef->size / Hugepagesize;

                        _qemuRetrySetPagePoolSize(cfg, def, cell,
                            HP_TARGET, Hugepagesize, pageinfo[PAGE_2M].avail - GuestNodeSize, false);
                    }
                }

                def->acquiredHugepages = false;
                err = 0;
            }
        }

 error:
        /* UNLOCK domain hugepage acquisition. */
        virMutexUnlock(&driver->hugepage_lock);

        VIR_FREE(pageinfo);

        return err;
    }

    return 0;
}
#endif /* defined(ENABLE_EXADATA) */
